import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            np.array([5], dtype="timedelta64[s]").reshape(1, 1),
            "dur_s.mat",
            "dur1",
        ),
        (
            np.array([5], dtype="timedelta64[m]").reshape(1, 1),
            "dur_m.mat",
            "dur2",
        ),
        (
            np.array([5], dtype="timedelta64[h]").reshape(1, 1),
            "dur_h.mat",
            "dur3",
        ),
        (
            np.array([5], dtype="timedelta64[D]").reshape(1, 1),
            "dur_d.mat",
            "dur4",
        ),
        (
            np.array([1, 2, 3], dtype="timedelta64[Y]").reshape(1, 3),
            "dur_y.mat",
            "dur8",
        ),
        (
            (np.timedelta64(1, "h") + np.timedelta64(2, "m") + np.timedelta64(3, "s"))
            .astype("timedelta64[ms]")
            .reshape(1, 1),
            "dur_base.mat",
            "dur5",
        ),
        (
            np.array([10, 20, 30, 40, 50, 60], dtype="timedelta64[s]").reshape(2, 3),
            "dur_array.mat",
            "dur6",
        ),
        (
            np.array([], dtype="datetime64[ms]"),
            "dur_empty.mat",
            "dur7",
        ),
    ],
    ids=[
        "duration-seconds",
        "duration-minutes",
        "duration-hours",
        "duration-days",
        "duration-years",
        "duration-base",
        "duration-array",
        "duration-empty",
    ],
)
def test_load_duration(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)



import os

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_df, file_name, var_name",
    [
        (
            pd.DataFrame({"Var1": [1.1, 2.2, 3.3], "Var2": [4.4, 5.5, 6.6]}),
            "T1.mat",
            "T1",
        ),
        (
            pd.DataFrame(
                {
                    "Var1": ["apple", "banana", "cherry"],
                }
            ),
            "T2.mat",
            "T2",
        ),
        (
            pd.DataFrame(
                {
                    "Time": np.array(
                        [
                            "2020-01-01T00:00:00.000",
                            "2020-01-02T00:00:00.000",
                            "2020-01-03T00:00:00.000",
                        ],
                        dtype="datetime64[ms]",
                    ),
                    "Duration": np.array([30, 60, 90], dtype="timedelta64[s]"),
                }
            ),
            "T3.mat",
            "T3",
        ),
        (
            pd.DataFrame(
                {
                    "Var1": [
                        np.array([[1.0]]),
                        np.array(["text"]),
                        np.array([["2023-01-01T00:00:00.000"]], dtype="datetime64[ms]"),
                    ],
                }
            ),
            "T5.mat",
            "T5",
        ),
        (
            pd.DataFrame(
                {
                    "Var1": [1.1, np.nan, 3.3],
                    "Var2": np.array(["A", "", "C"]),
                }
            ),
            "T6.mat",
            "T6",
        ),
        (
            pd.DataFrame(
                {
                    "X": np.array([10.0, 20.0, 30.0]),
                    "Y": np.array([100.0, 200.0, 300.0]),
                }
            ),
            "T7.mat",
            "T7",
        ),
        (
            pd.DataFrame(
                {
                    "Col1": np.array([1.0, 2.0, 3.0]),
                    "Col2": np.array([4.0, 5.0, 6.0]),
                },
                index=["R1", "R2", "R3"],
            ),
            "T8.mat",
            "T8",
        ),
        (
            pd.DataFrame(
                {
                    "Var1": np.array(
                        [
                            "2023-01-01T00:00:00.000",
                            "2023-01-02T00:00:00.000",
                            "2023-01-03T00:00:00.000",
                        ],
                        dtype="datetime64[ms]",
                    ),
                    "data_1": np.array([1.0, 2.0, 3.0]),
                    "data_2": np.array([4.0, 5.0, 6.0]),
                }
            ),
            "T10.mat",
            "T10",
        ),
    ],
    ids=[
        "simple-table",
        "table-with-strings",
        "table-with-datetime",
        "table-with-mixed-types",
        "table-with-nan",
        "table-with-row-names",
        "table-with-row-names",
        "table-with-multicolumn-var",
    ],
)
def test_parse_table(expected_df, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


def make_cell_entry(sub_arr):
    arr = np.empty(3, dtype=object)
    arr[:] = [sub_arr, sub_arr, sub_arr]
    return arr


@pytest.mark.parametrize(
    "expected_df, file_name, var_name",
    [
        (
            pd.DataFrame(
                {
                    "C": make_cell_entry(
                        np.array(
                            [[(np.array([[123.0]]), np.array(["abc"]))]],
                            dtype=[("field1", "O"), ("field2", "O")],
                        )
                    ),
                    "Var2": make_cell_entry(
                        {
                            "_Class": "MyObj",
                            "_Props": np.array(
                                [[{"Value": np.array([[42.0]])}]], dtype=object
                            ),
                        }
                    ),
                }
            ),
            "T4.mat",
            "T4",
        )
    ],
    ids=["table-with-struct-and-object"],
)
def test_parse_table_struct_and_object(expected_df, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)
    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, file_name, var_name",
    [
        (
            pd.DataFrame(
                {
                    "ID": np.array([1.0, 2.0]),
                    "Label": np.array(["one", "two"]),
                }
            ),
            "T9.mat",
            "T9",
        ),
    ],
    ids=["table-with-attrs"],
)
def test_parse_table_with_attrs(expected_df, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False, add_table_attrs=True)
    expected_df.attrs = {
        "Description": "Test table with full metadata",
        "DimensionNames": ["RowId", "Features"],
        "VariableUnits": ["", "category"],
        "VariableDescriptions": ["ID number", "Category label"],
        "VariableContinuity": ["continuous", "step"],
        "UserData": np.array(
            [[(np.array(["UnitTest"]), np.array([[1.0]]))]],
            dtype=[("CreatedBy", "O"), ("Version", "O")],
        ),
    }

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)

    # Check attributes
    for key, value in expected_df.attrs.items():
        assert key in matdict[var_name].attrs
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(matdict[var_name].attrs[key], value)
        else:
            assert matdict[var_name].attrs[key] == value



import os

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_df, file_name, var_name",
    [
        (
            pd.DataFrame(
                {"data1": [1.0, 2.0, 3.0]},
                index=pd.Index(
                    np.array(
                        ["2023-01-01", "2023-01-02", "2023-01-03"],
                        dtype="datetime64[ms]",
                    ),
                    name="Time",
                ),
            ),
            "tt1.mat",
            "tt1",
        ),
        (
            pd.DataFrame(
                {
                    "data2_1": [1.0, 2.0, 3.0],
                    "data2_2": [4.0, 5.0, 6.0],
                },
                index=pd.Index(
                    np.array([10, 20, 30], dtype="timedelta64[s]"),
                    name="Time",
                ),
            ),
            "tt2.mat",
            "tt2",
        ),
        (
            pd.DataFrame(
                {
                    "data1": [1.0, 2.0, 3.0],
                    "data3": [7.0, 8.0, 9.0],
                },
                index=pd.Index(
                    np.array(
                        ["2023-01-01", "2023-01-02", "2023-01-03"],
                        dtype="datetime64[ms]",
                    ),
                    name="Time",
                ),
            ),
            "tt3.mat",
            "tt3",
        ),
        (
            pd.DataFrame(
                {
                    "data1": [1.0, 2.0, 3.0],
                },
                index=pd.Index(
                    np.array([0, int(1e5), int(2e5)], dtype="timedelta64[ns]"),
                    name="Time",
                ),
            ),
            "tt4.mat",
            "tt4",
        ),
        (
            pd.DataFrame(
                {
                    "data1": [1.0, 2.0, 3.0],
                },
                index=pd.Index(
                    np.array([0, int(1e9), int(2e9)], dtype="timedelta64[ns]"),
                    name="Time",
                ),
            ),
            "tt5.mat",
            "tt5",
        ),
        (
            pd.DataFrame(
                {
                    "data1": [1.0, 2.0, 3.0],
                },
                index=pd.Index(
                    np.array(
                        [int(10e9), int(11e9), int(12e9)], dtype="timedelta64[ns]"
                    ),
                    name="Time",
                ),
            ),
            "tt6.mat",
            "tt6",
        ),
        (
            pd.DataFrame(
                {
                    "data1": [1.0, 2.0, 3.0],
                },
                index=pd.Index(
                    np.array(
                        [
                            "2020-01-01T00:00:00",
                            "2020-01-01T00:00:01",
                            "2020-01-01T00:00:02",
                        ],
                        dtype="datetime64[ns]",
                    ),
                    name="Time",
                ),
            ),
            "tt7.mat",
            "tt7",
        ),
        (
            pd.DataFrame(
                {"Pressure": [1.0, 2.0, 3.0]},
                index=pd.Index(
                    np.array(
                        ["2023-01-01", "2023-01-02", "2023-01-03"],
                        dtype="datetime64[ms]",
                    ),
                    name="Time",
                ),
            ),
            "tt8.mat",
            "tt8",
        ),
        (
            pd.DataFrame(
                {
                    "data1": [1.0, 2.0, 3.0],
                },
                index=pd.Index(
                    np.array(
                        [
                            "2020-01-01T00:00:00",
                            "2020-04-01T00:00:01",
                            "2020-07-01T00:00:02",
                        ],
                        dtype="datetime64[M]",
                    ),
                    name="Time",
                ),
            ),
            "tt10.mat",
            "tt10",
        ),
    ],
    ids=[
        "simple-timetable-datetime",
        "simple-timetable-datetime-multicolumn",
        "simple-timetable-datetime-multivars",
        "timetable-samplerate",
        "timetable-timestep",
        "timetable-samplerate-starttime",
        "timetable-timestep-starttime",
        "timetable-varnames",
        "timetable-calendarDuration-timestep",
    ],
)
def test_parse_timetable(expected_df, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)


@pytest.mark.parametrize(
    "expected_df, file_name, var_name",
    [
        (
            pd.DataFrame(
                {"data1": [1.0, 2.0, 3.0]},
                index=pd.Index(
                    np.array(
                        ["2023-01-01", "2023-01-02", "2023-01-03"],
                        dtype="datetime64[ms]",
                    ),
                    name="Date",
                ),
            ),
            "tt9.mat",
            "tt9",
        ),
    ],
    ids=["timetable-with-attrs"],
)
def test_parse_table_with_attrs(expected_df, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False, add_table_attrs=True)
    expected_df.attrs = {
        "Description": "Random Description",
        "varUnits": ["m/s"],
        "varDescriptions": ["myVar"],
        "varContinuity": ["continuous"],
        "UserData": np.empty((0, 0), dtype=float),
    }

    assert var_name in matdict
    pd.testing.assert_frame_equal(matdict[var_name], expected_df)

    # Check attributes
    for key, value in expected_df.attrs.items():
        assert key in matdict[var_name].attrs
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(matdict[var_name].attrs[key], value)
        else:
            assert matdict[var_name].attrs[key] == value



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            (
                np.empty((0, 0), dtype="timedelta64[M]"),
                np.empty((0, 0), dtype="timedelta64[D]"),
                np.empty((0, 0), dtype="timedelta64[ms]"),
            ),
            "cdur1.mat",
            "cdur1",
        ),
        (
            (
                np.array([[0]], dtype="timedelta64[M]"),
                np.array([[1, 2, 3]], dtype="timedelta64[D]"),
                np.array([[0]], dtype="timedelta64[ms]"),
            ),
            "cdur2.mat",
            "cdur2",
        ),
        (
            (
                np.array([[0]], dtype="timedelta64[M]"),
                np.array([[7, 14]], dtype="timedelta64[D]"),
                np.array([[0]], dtype="timedelta64[ms]"),
            ),
            "cdur3.mat",
            "cdur3",
        ),
        (
            (
                np.array([[1, 0]], dtype="timedelta64[M]"),
                np.array([[1, 2]], dtype="timedelta64[D]"),
                np.array([[0]], dtype="timedelta64[ms]"),
            ),
            "cdur4.mat",
            "cdur4",
        ),
        (
            (
                np.array([[12, 18]], dtype="timedelta64[M]"),
                np.array([[0]], dtype="timedelta64[D]"),
                np.array([[0]], dtype="timedelta64[ms]"),
            ),
            "cdur5.mat",
            "cdur5",
        ),
        (
            (
                np.array([[3]], dtype="timedelta64[M]"),
                np.array([[15]], dtype="timedelta64[D]"),
                np.array([[0]], dtype="timedelta64[ms]"),
            ),
            "cdur6.mat",
            "cdur6",
        ),
        (
            (
                np.array([[1, 0], [2, 0]], dtype="timedelta64[M]"),
                np.array([[0, 5], [0, 10]], dtype="timedelta64[D]"),
                np.array([[0, 0], [0, 0]], dtype="timedelta64[ms]"),
            ),
            "cdur7.mat",
            "cdur7",
        ),
        (
            (
                np.array([[0]], dtype="timedelta64[M]"),
                np.array([[1]], dtype="timedelta64[D]"),
                np.array([[3723000]], dtype="timedelta64[ms]"),
            ),
            "cdur8.mat",
            "cdur8",
        ),
    ],
    ids=[
        "calendarDuration-empty",
        "calendarDuration-days",
        "calendarDuration-weeks",
        "calendarDuration-mixed",
        "calendarDuration-years",
        "calendarDuration-quarters",
        "calendarDuration-2D",
        "calendarDuration-duration",
    ],
)
def test_load_duration(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    actual_tup = matdict[var_name][0, 0]["calendarDuration"]
    assert np.array_equal(actual_tup[0], expected_array[0])
    assert np.array_equal(actual_tup[1], expected_array[1])
    assert np.array_equal(actual_tup[2], expected_array[2])



"""Reads MAT-files and extracts variables including MATLAB objects"""

from io import BytesIO

import numpy as np
from scipy.io import loadmat
from scipy.io.matlab._mio5 import MatFile5Reader
from scipy.io.matlab._mio5_params import OPAQUE_DTYPE

from matio.subsystem import SubsystemReader


def get_matfile_version(ss_stream):
    """Reads subsystem MAT-file version and endianness
    Inputs
        1. ss_stream (BytesIO): Subsystem data stream
    Returns:
        1. v_major (int): Major version
        2. v_minor (int): Minor version
        3. byte_order (str): Endianness
    """

    ss_stream.seek(0)
    data = ss_stream.read(4)
    maj_ind = int(data[2] == b"I"[0])
    v_major = int(data[maj_ind])
    v_minor = int(data[1 - maj_ind])
    byte_order = "<" if data[2] == b"I"[0] else ">"
    if v_major in (1, 2):
        return v_major, v_minor, byte_order

    # Unsure if Subsystem and MAT-file versions can be different
    raise ValueError(f"Unknown subsystem data type, version {v_major}.{v_minor}")


def read_subsystem(ssdata, **kwargs):
    """Reads subsystem data as a MAT-file stream
    Inputs
        1. ssdata (numpy.ndarray): Subsystem data from "__function_workspace__"
        2. kwargs: Additional arguments for scipy.io.loadmat
    Returns:
        1. subsystem data (numpy.ndarray): Parsed subsystem data
        2. byte_order (str)
    """
    ss_stream = BytesIO(ssdata)

    mjv, mnv, byte_order = get_matfile_version(ss_stream)
    if mjv != 1:
        raise NotImplementedError(f"Subsystem version {mjv}.{mnv} not supported")

    kwargs.pop("byte_order", None)
    kwargs.pop("variable_names", None)

    ss_stream.seek(8)  # Skip subsystem header
    mat_reader = MatFile5Reader(ss_stream, byte_order=byte_order, **kwargs)
    mat_reader.initialize_read()
    hdr, _ = mat_reader.read_var_header()
    try:
        res = mat_reader.read_var_array(hdr, process=False)
    except Exception as err:
        raise ValueError(f"Error reading subsystem data: {err}") from err

    return res, byte_order


def remove_unsupported_args(kwargs):
    """Removes unsupported arguments for scipy.io.loadmat"""
    kwargs.pop("simplify_cells", None)
    kwargs.pop("squeeze_me", None)
    kwargs.pop("struct_as_record", None)
    kwargs.pop("uint16_codec", None)
    kwargs.pop("chars_as_strings", False)


def get_function_workspace(file_path, mdict=None, spmatrix=True, **kwargs):
    """Reads function workspace from MAT-file
    Inputs
        1. file_path (str): Path to MAT-file
        2. mdict (dict): Dictionary to store loaded variables
        3. spmatrix (bool): Whether to load sparse matrices
        4. kwargs: Additional arguments for scipy.io.loadmat
    Returns:
        1. matfile_dict (dict): Dictionary of loaded variables
        2. ssdata: Subsystem data from "__function_workspace__"
    """
    matfile_dict = loadmat(file_path, mdict=mdict, spmatrix=spmatrix, **kwargs)
    ssdata = matfile_dict.pop("__function_workspace__", None)
    return matfile_dict, ssdata


def find_opaque_dtype(arr, subsystem, path=()):
    """Recursively finds and replaces mxOPAQUE_CLASS objects in a numpy array
    with the corresponding MCOS object.

    This is a hacky solution to find mxOPAQUE_CLASS arrays inside struct arrays or cell arrays.
    """

    if not isinstance(arr, np.ndarray):
        return arr

    if arr.dtype == OPAQUE_DTYPE:
        type_system = arr[0]["_TypeSystem"]
        metadata = arr[0]["_Metadata"]
        return subsystem.read_mcos_object(metadata, type_system)

    if arr.dtype == object:
        # Iterate through cell arrays
        for idx in np.ndindex(arr.shape):
            cell_item = arr[idx]
            if cell_item.dtype == OPAQUE_DTYPE:
                type_system = cell_item[0]["_TypeSystem"]
                metadata = cell_item[0]["_Metadata"]
                arr[idx] = subsystem.read_mcos_object(metadata, type_system)
            else:
                find_opaque_dtype(cell_item, subsystem, path + (idx,))

    elif arr.dtype.names:
        # Iterate though struct array
        for idx in np.ndindex(arr.shape):
            for name in arr.dtype.names:
                field_val = arr[idx][name]
                if field_val.dtype == OPAQUE_DTYPE:
                    type_system = field_val[0]["_TypeSystem"]
                    metadata = field_val[0]["_Metadata"]
                    arr[idx][name] = subsystem.read_mcos_object(metadata, type_system)
                else:
                    find_opaque_dtype(field_val, subsystem, path + (idx, name))

    return arr


def load_from_mat(
    file_path,
    mdict=None,
    raw_data=False,
    add_table_attrs=False,
    *,
    spmatrix=True,
    **kwargs,
):
    """Loads variables from MAT-file
    Calls scipy.io.loadmat to read the MAT-file and then processes the
    "__function_workspace__" variable to extract subsystem data.
    Inputs
        1. file_path (str): Path to MAT-file
        2. mdict (dict): Dictionary to store loaded variables
        3. raw_data (bool): Whether to return raw data for objects
        4. add_table_attrs (bool): Add attributes to pandas DataFrame for MATLAB tables/timetables
        5. spmatrix (bool): Additional arguments for scipy.io.loadmat
        6. kwargs: Additional arguments for scipy.io.loadmat
    Returns:
        1. mdict (dict): Dictionary of loaded variables
    """

    remove_unsupported_args(kwargs)

    matfile_dict, ssdata = get_function_workspace(file_path, mdict, spmatrix, **kwargs)
    if ssdata is None:
        # No subsystem data in file
        if mdict is not None:
            mdict.update(matfile_dict)
            return mdict
        return matfile_dict

    ss_array, byte_order = read_subsystem(ssdata, **kwargs)
    subsystem = SubsystemReader(ss_array, byte_order, raw_data, add_table_attrs)

    for var, data in matfile_dict.items():
        if not isinstance(data, np.ndarray):
            continue
        matfile_dict[var] = find_opaque_dtype(data, subsystem)

    # Update mdict if present
    if mdict is not None:
        mdict.update(matfile_dict)
    else:
        mdict = matfile_dict

    return mdict



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "categories, codes, ordered, file_name, var_name",
    [
        (
            np.array(["blue", "green", "red"]),
            np.array([[2, 1, 0, 2]]),
            False,
            "cat1.mat",
            "cat1",
        ),
        (
            np.array(["high", "low", "medium"]),
            np.array([[1, 2], [0, 1]]),
            False,
            "cat2.mat",
            "cat2",
        ),
        (
            np.array(["cold", "warm", "hot"]),
            np.array([[0, 2, 1]]),
            False,
            "cat3.mat",
            "cat3",
        ),
        (
            np.array(["small", "medium", "large"]),
            np.array([[0, 1, 2]]),
            True,
            "cat4.mat",
            "cat4",
        ),
        (
            np.array(["low", "medium", "high"]),
            np.array([[0, 1, 2, 1, 0]]),
            False,
            "cat5.mat",
            "cat5",
        ),
        (
            [],
            np.empty((0, 0), dtype=np.uint8),
            False,
            "cat6.mat",
            "cat6",
        ),
        (
            np.array(["cat", "dog", "mouse"]),
            np.array([[0, -1, 1, 2]]),
            False,
            "cat7.mat",
            "cat7",
        ),
        (
            np.array(["autumn", "spring", "summer", "winter"]),
            np.array([[1, 2, 0, 3]]),
            False,
            "cat8.mat",
            "cat8",
        ),
        (
            np.array(["OFF", "ON", "On", "off", "on"]),
            np.array([[2, 3, 0, 1, 4]]),
            False,
            "cat9.mat",
            "cat9",
        ),
        (
            np.array(["maybe", "no", "yes"]),
            np.tile(np.array([[2, 2], [1, 1], [0, 0]], dtype=np.uint8), (2, 1, 1)),
            False,
            "cat10.mat",
            "cat10",
        ),
    ],
    ids=[
        "categorical-basic",
        "categorical-2D",
        "categorical-explicit-cats",
        "categorical-ordered",
        "categorical-numeric-labels",
        "categorical-empty",
        "categorical-missing-labels",
        "categorical-from-string",
        "categorical-mixed-case-cats",
        "categorical-3D",
    ],
)
def test_load_categorical(categories, codes, ordered, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)
    assert var_name in matdict
    assert np.array_equal(matdict[var_name].codes, codes)
    assert np.array_equal(matdict[var_name].categories, categories)
    assert np.array_equal(matdict[var_name].ordered, ordered)



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            [
                (np.float64(1), np.str_("apple")),
                (np.float64(2), np.str_("banana")),
                (np.float64(3), np.str_("cherry")),
            ],
            "dict1.mat",
            "dict1",
        ),
        (
            [
                (np.str_("x"), np.float64(10)),
                (np.str_("y"), np.float64(20)),
                (np.str_("z"), np.float64(30)),
            ],
            "dict2.mat",
            "dict2",
        ),
        (
            [
                (np.str_("name"), np.array([["Alice"]])),
                (np.str_("age"), np.array([[25]])),
            ],
            "dict3.mat",
            "dict3",
        ),
        (
            [
                (np.array([[1]]), np.str_("one")),
                (np.array([[2]]), np.str_("two")),
                (np.array([[3]]), np.str_("three")),
            ],
            "dict4.mat",
            "dict4",
        ),
    ],
    ids=[
        "dict-numeric-key",
        "dict-char-key",
        "dict-mixed-val",
        "dict-cell-key",
    ],
)
def test_load_containermap(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    for i, (expected_key, expected_val) in enumerate(expected_array):
        actual_key = matdict[var_name][i][0]
        actual_val = matdict[var_name][i][1]
        assert np.array_equal(actual_key, expected_key)
        assert np.array_equal(actual_val, expected_val)



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            {
                "_Class": "containers.Map",
                "_Props": {
                    1: np.array(["a"]),
                    2: np.array(["b"]),
                },
            },
            "map2.mat",
            "map2",
        ),
        (
            {
                "_Class": "containers.Map",
                "_Props": {
                    "a": np.array([[1]]),
                    "b": np.array([[2]]),
                },
            },
            "map3.mat",
            "map3",
        ),
        (
            {
                "_Class": "containers.Map",
                "_Props": {
                    "a": np.array([[1]]),
                    "b": np.array([[2]]),
                },
            },
            "map4.mat",
            "map4",
        ),
    ],
    ids=[
        "map-numeric-key",
        "map-char-key",
        "map-string-key",
    ],
)
def test_load_containermap(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    assert matdict[var_name]["_Class"] == expected_array["_Class"]
    for key in expected_array["_Props"]:
        assert key in matdict[var_name]["_Props"]
        assert np.array_equal(
            matdict[var_name]["_Props"][key], expected_array["_Props"][key]
        )


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            {
                "_Class": "containers.Map",
                "_Props": {},
            },
            "map1.mat",
            "map1",
        ),
    ],
    ids=[
        "map-empty",
    ],
)
def test_load_empty_containermap(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    assert matdict[var_name]["_Class"] == expected_array["_Class"]
    assert matdict[var_name]["_Props"] == expected_array["_Props"]



"""Utility functions for convertin MATLAB strings"""

import warnings

import numpy as np


def mat_to_string(props, byte_order):
    """Parse string data from MATLAB file
    String objects are stored as "any" properties in MAT-files.

    Strings are stored within a uint64 array with the following format:
        1. version
        2. ndims
        3. shape
        4. char_counts
        5. List of null-terminated strings as uint16 integers
    """

    data = props[0, 0].get("any", np.array([]))
    if data.size == 0:
        return np.array([[]], dtype=np.str_)

    if data[0, 0] != 1:
        warnings.warn(
            "String saved from a different MAT-file version. This may work unexpectedly",
            UserWarning,
        )

    ndims = data[0, 1]
    shape = data[0, 2 : 2 + ndims]
    num_strings = np.prod(shape)
    char_counts = data[0, 2 + ndims : 2 + ndims + num_strings]
    byte_data = data[0, 2 + ndims + num_strings :].tobytes()

    strings = []
    pos = 0
    encoding = "utf-16-le" if byte_order[0] == "<" else "utf-16-be"
    for char_count in char_counts:
        byte_length = char_count * 2  # UTF-16 encoding
        extracted_string = byte_data[pos : pos + byte_length].decode(encoding)
        strings.append(np.str_(extracted_string))
        pos += byte_length

    return np.reshape(strings, shape, order="F")



"""Reads MCOS subsystem data from MAT files"""

import warnings

import numpy as np

from matio.convert import convert_to_object, mat_to_enum


class SubsystemReader:
    """Extracts object properties from the subsystem data
    Currently only supports MCOS objects
    """

    def __init__(self, ssdata, byte_order, raw_data=False, add_table_attrs=False):
        self.byte_order = (
            "<u4" if byte_order == "<" else ">u4"
        )  # Could potentially be int32
        self.raw_data = raw_data
        self.add_table_attrs = add_table_attrs
        self.fwrap_metadata = None
        self.fwrap_vals = None
        self.fwrap_defaults = None
        self.mcos_names = None
        self.MAX_OBJECT_ID = -1
        self.MAX_CLASS_ID = -1
        self.init_fields(ssdata)

    def init_fields(self, ssdata):
        """Fetches metadata and field contents from the subsystem data
        Currently only supports MCOS objects
        Attributes:
            1. fwrap_metadata: Metadata for MCOS objects
            2. fwrap_vals: Numpy array of properties of MCOS objects
            3. fwrap_defaults: Numpy array of default properties of MCOS classes
            4. mcos_names: List of field and class names of all MCOS objects in file
        """
        if "MCOS" in ssdata.dtype.names:
            fwrap_data = ssdata[0, 0]["MCOS"][0]["_Metadata"]
            self.fwrap_metadata = fwrap_data[0, 0][:, 0]
            toc_flag = np.frombuffer(
                self.fwrap_metadata, dtype=self.byte_order, count=1, offset=0
            )[0]
            if toc_flag in (2, 3):
                # Not sure
                fwrap_version_offsets = 6
            elif toc_flag == 4:
                fwrap_version_offsets = 8
            else:
                raise ValueError("Incompatible FileWrapper version")

            self.fwrap_vals = fwrap_data[2:-3, 0]
            self.fwrap_defaults = fwrap_data[-3:, 0]
            self.mcos_names = self.get_field_names(fwrap_version_offsets)
            self.MAX_OBJECT_ID, self.MAX_CLASS_ID = self.get_max_ids()

    def get_max_ids(self):
        """Extracts the maximum object and class IDs from the metadata"""

        # Get offsets
        cid_start, cid_end, oid_start, oid_end = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=4, offset=8
        )

        max_class_id = (cid_end - cid_start) // 16 - 1
        max_object_id = (oid_end - oid_start) // 24 - 1
        return max_object_id, max_class_id

    def get_field_names(self, num_offsets):
        """Extracts field and class names from the subsystem data
        Names are stored as a list of null-terminated strings
        Inputs:
            1. num_offsets: Number of offsets in the metadata
            This is determined by FileWrapper version
        Returns:
            1. all_names: List of field and class names
        """
        byte_end = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=1, offset=8
        )[0]
        byte_start = 8 + num_offsets * 4
        data = self.fwrap_metadata[byte_start:byte_end].tobytes()
        raw_strings = data.split(b"\x00")
        all_names = [s.decode("ascii") for s in raw_strings if s]
        return all_names

    def get_object_dependencies(self, object_id):
        """Extracts object dependency IDs for a given object
        Dependency IDs are stored in blocks of 24 bytes ordered by object ID
        Each block contains:
            1. Class ID
            2. Unknown flag
            3. Unknown flag
            4. Type1 ID
            5. Type2 ID
            6. Dependency ID

        Inputs:
            1. object_id: ID of the object
        Returns:
            (class_id, type1_id, type2_id, dep_id)
        """

        byte_offset = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=1, offset=16
        )[0]
        byte_offset = byte_offset + object_id * 24
        class_id, _, _, type1_id, type2_id, dep_id = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=6, offset=byte_offset
        )

        return class_id, type1_id, type2_id, dep_id

    def get_class_name(self, class_id):
        """Extracts class name and handle for a given object from its class ID
        Class IDs are stored in blocks of 16 bytes ordered by class ID
        Each block contains:
            1. Handle Name Index
            2. Class Name Index
            3. Unknown flag
            4. Unknown flag
        Inputs:
            1. class_id: ID of the class
        Returns:
            (handle_name, class_name)
        """

        byte_offset = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=1, offset=8
        )[0]
        byte_offset = byte_offset + class_id * 16

        handle_idx, class_idx, _, _ = np.frombuffer(
            self.fwrap_metadata,
            dtype=self.byte_order,
            count=4,
            offset=byte_offset,
        )

        class_name = self.mcos_names[class_idx - 1]
        handle_name = self.mcos_names[handle_idx - 1] if handle_idx > 0 else None
        return handle_name, class_name

    def get_ids(self, m_id, byte_offset, nbytes):
        """Extract nblocks and subblock contents for a given object
        Helper method to parse metadata with the format:
            1. nsubblocks (4 bytes)
            2. subblock contents (nblocks * nbytes)

        Inputs:
            1. m_id: ID of the object
            2. byte_offset: Offset to start reading from
            3. nbytes: Number of bytes in each subblock
        Returns:
            1. ids: Numpy array of all subblock contents
        """

        # Get block corresponding to type ID
        while m_id > 0:
            nblocks = np.frombuffer(
                self.fwrap_metadata, dtype=self.byte_order, count=1, offset=byte_offset
            )[0]
            byte_offset = byte_offset + 4 + nblocks * nbytes
            if ((nblocks * nbytes) + 4) % 8 != 0:
                byte_offset += 4
            m_id -= 1

        # Get the number of blocks
        nblocks = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=1, offset=byte_offset
        )[0]

        byte_offset += 4
        ids = np.frombuffer(
            self.fwrap_metadata,
            dtype=self.byte_order,
            count=nblocks * nbytes // 4,
            offset=byte_offset,
        )

        return ids.reshape((nblocks, nbytes // 4))

    def get_handle_class_instance(self, type2_id):
        """Reads handle class instance ID for a given object
        Searches for the object ID of the corresponding handle class from its type 2 ID
        Inputs:
            1. type2_id: ID of the handle instance
        Returns:
            1. class_id of the handle instance
            2. object_id of the handle instance
        """

        start, end = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=2, offset=16
        )
        blocks = np.frombuffer(
            self.fwrap_metadata[start:end], dtype=self.byte_order
        ).reshape(-1, 6)

        for idx, block in enumerate(blocks):
            if block[4] == type2_id:
                class_id = block[0]
                object_id = idx
                return class_id, np.array([object_id])

        raise ValueError(f"Handle class instance not found for type2_id: {type2_id}")

    def extract_handles(self, dep_id):
        """Extracts the handle instances for an object
        Handle instances attached to an object are tagged by their type 2 IDs
        Objects with handle instances can be found by their dependency ID
        """

        # Get block corresponding to dep_id
        byte_offset = np.frombuffer(
            self.fwrap_metadata, dtype=self.byte_order, count=1, offset=24
        )[0]
        handle_type2_ids = self.get_ids(dep_id, byte_offset, nbytes=4)[:, 0]
        if handle_type2_ids.size == 0:
            return None

        handles = {}
        for i, handle_id in enumerate(handle_type2_ids):
            class_id, object_id = self.get_handle_class_instance(handle_id)
            handles[f"_Handle_{i + 1}"] = self.read_object_arrays(
                object_id, class_id, dims=[1, 1]
            )
        return handles

    def find_object_reference(self, arr, path=()):
        """Recursively searches for object references in the data array
        and replaces them with the corresponding MCOS object.

        This is a hacky solution to find object arrays inside struct arrays or cell arrays.
        """

        if not isinstance(arr, np.ndarray):
            return arr

        if check_object_reference(arr, self.MAX_OBJECT_ID, self.MAX_CLASS_ID):
            return self.read_mcos_object(arr)

        if arr.dtype == object:
            # Iterate through cell arrays
            for idx in np.ndindex(arr.shape):
                cell_item = arr[idx]
                if check_object_reference(
                    cell_item, self.MAX_OBJECT_ID, self.MAX_CLASS_ID
                ):
                    arr[idx] = self.read_mcos_object(cell_item)
                else:
                    self.find_object_reference(cell_item, path + (idx,))
                # Path to keep track of the current index
        elif arr.dtype.names:
            # Iterate through struct array
            if check_object_reference(arr, self.MAX_OBJECT_ID, self.MAX_CLASS_ID):
                return self.read_mcos_object(arr)
            for idx in np.ndindex(arr.shape):
                for name in arr.dtype.names:
                    field_val = arr[idx][name]
                    if check_object_reference(
                        field_val, self.MAX_OBJECT_ID, self.MAX_CLASS_ID
                    ):
                        arr[idx][name] = self.read_mcos_object(field_val)
                    else:
                        self.find_object_reference(field_val, path + (idx, name))

        return arr

    def parse_field_types(self, field_type, field_value, field_idx):
        """Parses field types and values for an object"""

        if field_type == 0:
            val = self.mcos_names[field_value - 1]
        elif field_type == 1:
            val = self.find_object_reference(self.fwrap_vals[field_value])
        elif field_type == 2:
            if field_value < 0 or field_value > 1:
                print("Field Value Found!", field_idx, field_type, field_value)
            val = field_value
        else:
            raise ValueError(f"Unknown field type: {field_type}")

        return val

    def extract_fields(self, object_id):
        """Extracts the properties for an object
        Inputs:
            (type1_id, type2_id, dep_id): Dependency IDs of the object
        Returns:
            1. obj_props: Dictionary of object properties keyed by property names
        """

        _, type1_id, type2_id, dep_id = self.get_object_dependencies(object_id)

        if type1_id == 0 and type2_id != 0:
            obj_type_id = type2_id
            byte_offset = np.frombuffer(
                self.fwrap_metadata, dtype=self.byte_order, count=1, offset=20
            )[0]
        elif type1_id != 0 and type2_id == 0:
            obj_type_id = type1_id
            byte_offset = np.frombuffer(
                self.fwrap_metadata, dtype=self.byte_order, count=1, offset=12
            )[0]
        else:
            raise ValueError("Could not determine object type")

        obj_props = {}
        field_ids = self.get_ids(obj_type_id, byte_offset, nbytes=12)
        for field_idx, field_type, field_value in field_ids:
            obj_props[self.mcos_names[field_idx - 1]] = self.parse_field_types(
                field_type, field_value, field_idx
            )

        # Include Handle Values
        handles = self.extract_handles(dep_id)
        if handles is not None:
            obj_props.update(handles)
        return obj_props

    def read_object_arrays(self, object_ids, class_id, dims):
        """Reads an object array for a given variable
        Inputs:
            1. object_ids: IDs of the objects
            2. class_id: ID of the class
            3. dims: Dimensions of the object array
        Returns:
            1. result: Dictionary representing the object array
            Dictionary contains:
                - _Class: Class name
                - _Props: Numpy array of object properties
        """

        # Attach class name to the object
        handle_name, class_name = self.get_class_name(class_id)
        if handle_name is not None:
            class_name = f"{handle_name}.{class_name}"

        if object_ids.size == 0:
            return {
                "_Class": class_name,
                "_Props": np.empty((0, 0), dtype=object),
            }

        props_list = []
        for object_id in object_ids:
            obj_props = self.extract_fields(object_id)
            props_list.append(obj_props)
        obj_props = np.array(props_list).reshape(dims)

        obj_default_props = self.fwrap_defaults[2][class_id, 0]
        obj_default_props = self.find_object_reference(obj_default_props)
        # Update object properties with any default values
        if obj_default_props.size != 0:
            for name in obj_default_props.dtype.names:
                default_val = obj_default_props[name][0, 0]
                for idx in np.ndindex(obj_props.shape):
                    if name not in obj_props[idx]:
                        obj_props[idx][name] = default_val

        # Converts some common MATLAB objects to Python objects
        result = convert_to_object(
            obj_props, class_name, self.byte_order, self.raw_data, self.add_table_attrs
        )

        # Remaining unknown class properties
        # _u1 = self.fwrap_defaults[0][class_id, 0]
        # _u2 = self.fwrap_defaults[1][class_id, 0]

        return result

    def read_mcos_enumeration(self, metadata):
        """Reads MCOS enumeration object from the metadata
        Inputs:
            metadata: Metadata for the enumeration object
        Returns:
            metadata: Metadata for the enumeration object
            If raw_data is False, returns the enumeration object
        """

        handle_name, class_name = self.get_class_name(
            metadata[0, 0]["ClassName"].item()
        )
        if handle_name is not None:
            class_name = f"{handle_name}.{class_name}"

        builtin_class_index = metadata[0, 0]["BuiltinClassName"].item()
        if builtin_class_index != 0:
            handle_name, builtin_class_name = self.get_class_name(builtin_class_index)
            if handle_name is not None:
                builtin_class_name = f"{handle_name}.{builtin_class_name}"
        else:
            builtin_class_name = None

        value_names = [
            self.mcos_names[val - 1] for val in metadata[0, 0]["ValueNames"].ravel()
        ]  # Array is N x 1 shape

        enum_vals = []
        value_idx = metadata[0, 0]["ValueIndices"]
        mmdata = metadata[0, 0]["Values"]  # Array is N x 1 shape
        if mmdata.size != 0:
            mmdata_map = mmdata[value_idx]
            for val in np.nditer(mmdata_map, flags=["refs_ok"], op_flags=["readonly"]):
                obj_array = self.read_normal_mcos(val.item())
                enum_vals.append(obj_array)

        if not self.raw_data:
            enum_array = mat_to_enum(
                enum_vals,
                value_names,
                class_name,
                value_idx.shape,
            )
            return enum_array

        metadata[0, 0]["BuiltinClassName"] = builtin_class_name
        metadata[0, 0]["ClassName"] = class_name
        metadata[0, 0]["ValueNames"] = np.array(value_names).reshape(
            value_idx.shape, order="F"
        )
        metadata[0, 0]["ValueIndices"] = value_idx
        metadata[0, 0]["Values"] = np.array(enum_vals).reshape(
            value_idx.shape, order="F"
        )

        return metadata

    def read_normal_mcos(self, metadata):
        """Reads normal MCOS object from the metadata"""

        ndims = metadata[1, 0]
        dims = metadata[2 : 2 + ndims, 0]
        total_objs = np.prod(np.array(dims))
        object_ids = metadata[2 + ndims : 2 + ndims + total_objs, 0]
        class_id = metadata[-1, 0]
        return self.read_object_arrays(object_ids, class_id, dims)

    def read_mcos_object(self, metadata, type_system="MCOS"):
        """Reads MCOS object based on OPAQUE_DTYPE CONTENTS"""
        if type_system != "MCOS":
            warnings.warn(
                f"Type system {type_system} is not supported.",
                UserWarning,
            )
            return metadata

        if metadata.dtype.names is not None:
            if "EnumerationInstanceTag" in metadata.dtype.names:
                return self.read_mcos_enumeration(metadata)

            warnings.warn(
                "Couldn't read MCOS object type, returning object metadata",
                UserWarning,
            )
            return metadata

        if metadata.dtype == np.uint32:
            return self.read_normal_mcos(metadata)

        return metadata


def check_enumeration_instance_tag(metadata):
    """Checks if the metadata contains an enumeration instance tag"""
    if metadata.dtype.names is not None:
        if "EnumerationInstanceTag" in metadata.dtype.names:
            if metadata[0, 0]["EnumerationInstanceTag"] == 0xDD000000:
                return True
    return False


def check_object_reference(metadata, max_object_id, max_class_id):
    """Checks if the metadata is a valid object reference"""
    # Not sure if they include checks for max_object_id/max_class_id
    if not isinstance(metadata, np.ndarray):
        return False

    if check_enumeration_instance_tag(metadata):
        return True

    if metadata.dtype != np.uint32 or metadata.size < 5:
        return False

    if len(metadata.shape) == 2 and metadata.shape[1] != 1:
        return False

    # Magic Value
    if metadata[0, 0] != 0xDD000000:
        return False

    ndims = metadata[1, 0]
    if ndims <= 1:
        return False

    dims = metadata[2 : 2 + ndims, 0]
    total_objs = np.prod(dims)
    if total_objs == 0:
        # Can be an empty 0D object array
        if metadata.shape[0] != 3 + ndims or metadata[-1, 0] > max_class_id:
            return False
        return True

    object_ids = metadata[2 + ndims : 2 + ndims + total_objs, 0]
    if np.any(object_ids <= 0) or np.any(object_ids > max_object_id):
        return False
    if object_ids.size + ndims + 3 != metadata.shape[0]:
        return False

    class_id = metadata[-1, 0]
    return 0 < class_id <= max_class_id



"""Convert MATLAB objects to Python compatible objects"""

from enum import Enum

import numpy as np

from matio.utils import (
    mat_to_calendarduration,
    mat_to_categorical,
    mat_to_containermap,
    mat_to_datetime,
    mat_to_dictionary,
    mat_to_duration,
    mat_to_string,
    mat_to_table,
    mat_to_timetable,
)


def convert_to_object(
    props, class_name, byte_order, raw_data=False, add_table_attrs=False
):
    """Converts the object to a Python compatible object"""

    if raw_data:
        return {
            "_Class": class_name,
            "_Props": props,
        }

    class_to_function = {
        "datetime": lambda: mat_to_datetime(props),
        "duration": lambda: mat_to_duration(props),
        "string": lambda: mat_to_string(props, byte_order),
        "table": lambda: mat_to_table(props, add_table_attrs),
        "timetable": lambda: mat_to_timetable(props, add_table_attrs),
        "containers.Map": lambda: {
            "_Class": class_name,
            "_Props": mat_to_containermap(props),
        },
        "categorical": lambda: mat_to_categorical(props),
        "dictionary": lambda: mat_to_dictionary(props),
        "calendarDuration": lambda: mat_to_calendarduration(props),
    }

    result = class_to_function.get(
        class_name,
        lambda: {"_Class": class_name, "_Props": props},  # Default case
    )()

    return result


def mat_to_enum(values, value_names, class_name, shapes):
    """Converts MATLAB enum to Python enum"""

    enum_class = Enum(
        class_name,
        {name: val["_Props"].item() for name, val in zip(value_names, values)},
    )

    enum_members = [enum_class(val["_Props"].item()) for val in values]
    return np.array(enum_members, dtype=object).reshape(shapes, order="F")



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            np.array([["2025-04-01T12:00:00"]], dtype="datetime64[ms]").reshape(1, 1),
            "dt_base.mat",
            "dt1",
        ),
        (
            np.array([["2025-04-01T12:00:00"]], dtype="datetime64[ms]").reshape(1, 1),
            "dt_tz.mat",
            "dt2",
        ),
        (
            np.array(
                [
                    [
                        "2025-04-01",
                        "2025-04-03",
                        "2025-04-05",
                        "2025-04-02",
                        "2025-04-04",
                        "2025-04-06",
                    ]
                ],
                dtype="datetime64[ms]",
            ).reshape(2, 3),
            "dt_array.mat",
            "dt3",
        ),
        (
            np.array([], dtype="datetime64[ms]"),
            "dt_empty.mat",
            "dt4",
        ),
    ],
    ids=[
        "datetime-basic",
        "datetime-timezone",
        "datetime-array",
        "datetime-empty",
    ],
)
def test_load_datetime(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)



"""Utility functions for converting MATLAB tables and timetables to pandas DataFrames"""

import warnings

import numpy as np
import pandas as pd

MAX_TABLE_VERSION = 4
MIN_TABLE_VERSION = 1
MAX_TIMETABLE_VERSION = 6
MIN_TIMETABLE_VERSION = 2


def add_table_props(df, tab_props):
    """Add MATLAB table properties to pandas DataFrame
    These properties are mostly cell arrays of character vectors
    """

    df.attrs["Description"] = (
        tab_props["Description"].item() if tab_props["Description"].size > 0 else ""
    )
    df.attrs["VariableDescriptions"] = [
        s.item() if s.size > 0 else ""
        for s in tab_props["VariableDescriptions"].ravel()
    ]
    df.attrs["VariableUnits"] = [
        s.item() if s.size > 0 else "" for s in tab_props["VariableUnits"].ravel()
    ]
    df.attrs["VariableContinuity"] = [
        s.item() if s.size > 0 else "" for s in tab_props["VariableContinuity"].ravel()
    ]
    df.attrs["DimensionNames"] = [
        s.item() if s.size > 0 else "" for s in tab_props["DimensionNames"].ravel()
    ]
    df.attrs["UserData"] = tab_props["UserData"]

    return df


def add_timetable_props(df, tab_props):
    """Add MATLAB table properties to pandas DataFrame
    These properties are mostly cell arrays of character vectors
    """
    df.attrs["varDescriptions"] = [
        s.item() if s.size > 0 else "" for s in tab_props["varDescriptions"].ravel()
    ]
    df.attrs["varUnits"] = [
        s.item() if s.size > 0 else "" for s in tab_props["varUnits"].ravel()
    ]
    df.attrs["varContinuity"] = [
        s.item() if s.size > 0 else "" for s in tab_props["varContinuity"].ravel()
    ]
    df.attrs["UserData"] = tab_props["arrayProps"]["UserData"][0, 0]
    df.attrs["Description"] = (
        tab_props["arrayProps"]["Description"][0, 0].item()
        if tab_props["arrayProps"]["Description"][0, 0].size > 0
        else ""
    )

    return df


def to_dataframe(data, nvars, varnames):
    """Creates a dataframe from coldata and column names"""
    rows = {}
    for i in range(nvars):
        vname = varnames[0, i].item()
        coldata = data[0, i]

        # If variable is multicolumn data
        if isinstance(coldata, np.ndarray):
            if coldata.shape[1] == 1:
                rows[vname] = coldata[:, 0]
            else:
                for j in range(coldata.shape[1]):
                    colname = f"{vname}_{j + 1}"
                    rows[colname] = coldata[:, j]
        else:
            rows[vname] = coldata

    df = pd.DataFrame(rows)
    return df


def mat_to_table(props, add_table_attrs=False):
    """Converts MATLAB table to pandas DataFrame"""

    ver = int(props[0, 0]["props"][0, 0]["versionSavedFrom"].item())
    if not MIN_TABLE_VERSION <= ver <= MAX_TABLE_VERSION:
        warnings.warn(
            f"MATLAB table version {ver} is not supported. "
            f"Minimum supported version is {MIN_TABLE_VERSION}.",
            UserWarning,
        )

    data = props[0, 0]["data"]
    nvars = int(props[0, 0]["nvars"].item())
    varnames = props[0, 0]["varnames"]
    df = to_dataframe(data, nvars, varnames)

    # Add df.index
    nrows = int(props[0, 0]["nrows"].item())
    rownames = props[0, 0]["rownames"]
    if rownames.size > 0:
        rownames = [s.item() for s in rownames.ravel()]
        if len(rownames) == nrows:
            df.index = rownames

    tab_props = props[0, 0]["props"][0, 0]
    if add_table_attrs:
        # Since pandas lists this as experimental, flag so we can switch off if it breaks
        df = add_table_props(df, tab_props)

    return df


def get_row_times(row_times, num_rows):
    """Get row times from MATLAB timetable
    rowTimes is a duration or datetime array if explicitly specified
    If using "SampleRate" or "TimeStep", it is a struct array with the following fields:
    1. origin - the start time as a duration or datetime scalar
    2. specifiedAsRate - boolean indicating which to use - sampleRate or TimeStep
    3. stepSize - the time step as a duration scalar
    4. sampleRate - the sample rate as a float
    """
    if not row_times.dtype.names:
        return row_times.ravel()

    start = row_times[0, 0]["origin"]
    if row_times[0, 0]["specifiedAsRate"]:
        fs = row_times[0, 0]["sampleRate"].item()
        step = np.timedelta64(int(1e9 / fs), "ns")
    else:
        step = row_times[0, 0]["stepSize"]
        if step.dtype.names is not None and "calendarDuration" in step.dtype.names:
            comps = step[0, 0]["calendarDuration"]
            step = comps[0] or comps[1] or comps[2]
            # Only one of months, days, or millis is non-zero
            step_unit = np.datetime_data(step.dtype)[0]
            start = start.astype(f"datetime64[{step_unit}]")
        else:
            step = step.astype("timedelta64[ns]")

    return (start + step * np.arange(num_rows)).ravel()


def mat_to_timetable(props, add_table_attrs=False):
    """Converts MATLAB timetable to pandas DataFrame"""

    ver = int(props[0, 0]["any"][0, 0]["versionSavedFrom"].item())
    if not MIN_TIMETABLE_VERSION <= ver <= MAX_TIMETABLE_VERSION:
        warnings.warn(
            f"MATLAB timetable version {ver} is not supported. "
            f"Minimum supported version is {MIN_TIMETABLE_VERSION}.",
            UserWarning,
        )

    num_vars = int(props[0, 0]["any"][0, 0]["numVars"].item())
    var_names = props[0, 0]["any"][0, 0]["varNames"]
    data = props[0, 0]["any"][0, 0]["data"]
    df = to_dataframe(data, num_vars, var_names)

    row_times = props[0, 0]["any"][0, 0]["rowTimes"]
    num_rows = int(props[0, 0]["any"][0, 0]["numRows"].item())

    row_times = get_row_times(row_times, num_rows)
    dim_names = props[0, 0]["any"][0, 0]["dimNames"]
    df.index = pd.Index(row_times, name=dim_names[0, 0].item())

    if add_table_attrs:
        # Since pandas lists this as experimental, flag so we can switch off if it breaks
        df = add_timetable_props(df, props[0, 0]["any"][0, 0])

    return df


def mat_to_categorical(props):
    """Converts MATLAB categorical to pandas Categorical
    MATLAB categorical objects are stored with the following properties:
    1. categoryNames - all unique categories
    2. codes
    3. isOrdinal - boolean indicating if the categorical is ordered
    4. isProtected - boolean indicating if the categorical is protected
    """

    raw_names = props[0, 0]["categoryNames"]
    category_names = [name.item() for name in raw_names.ravel()]

    # MATLAB codes are 1-indexed as uint integers
    codes = props[0, 0]["codes"].astype(int) - 1
    ordered = bool(props[0, 0]["isOrdinal"].item())
    return pd.Categorical.from_codes(codes, categories=category_names, ordered=ordered)



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (np.array([10]).reshape(1, 1), "var_int.mat", "var_int"),
    ],
    ids=["simple-string"],
)
def test_parse_no_object(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert matdict[var_name] == expected_array

    assert isinstance(matdict[var_name], np.ndarray)
    np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            np.array(
                [
                    np.array(["String in Cell"]).reshape(1, 1),
                ]
            ).reshape(1, 1),
            "var_cell.mat",
            "var_cell",
        ),
    ],
    ids=["string-in-cell"],
)
def test_parse_string_in_cell(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert isinstance(matdict[var_name], np.ndarray)
    np.testing.assert_array_equal(matdict[var_name], expected_array)


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            np.array(
                [[(np.array(["String in Struct"]).reshape(1, 1),)]],
                dtype=[("MyField", "O")],
            ),
            "var_struct.mat",
            "var_struct",
        )
    ],
    ids=["string-in-struct"],
)
def test_parse_string_in_struct(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)
    print(matdict[var_name])
    print(expected_array)
    # Output format
    assert var_name in matdict
    assert isinstance(matdict[var_name], np.ndarray)
    np.testing.assert_array_equal(matdict[var_name], expected_array)



import os
from enum import Enum

import numpy as np
import pytest

from matio import load_from_mat


class EnumClass(Enum):
    """Enum class for testing purposes."""

    enum1 = {"val": np.array([1.0]).reshape(1, 1)}
    enum2 = {"val": np.array([2.0]).reshape(1, 1)}
    enum3 = {"val": np.array([3.0]).reshape(1, 1)}
    enum4 = {"val": np.array([4.0]).reshape(1, 1)}
    enum5 = {"val": np.array([5.0]).reshape(1, 1)}
    enum6 = {"val": np.array([6.0]).reshape(1, 1)}


class EnumClass2(Enum):
    """Enum class for testing"""

    enum1 = {"uint32.Data": np.array([1], dtype=np.uint32).reshape(1, 1)}


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            np.array(
                [
                    [
                        EnumClass.enum1,
                    ]
                ]
            ),
            "enum_base.mat",
            "enum_base",
        ),
        (
            np.array(
                [
                    EnumClass.enum1,
                    EnumClass.enum2,
                    EnumClass.enum3,
                    EnumClass.enum4,
                    EnumClass.enum5,
                    EnumClass.enum6,
                ]
            ).reshape(2, 3, order="F"),
            "enum_array.mat",
            "enum_arr",
        ),
        (
            np.array(
                [
                    [
                        EnumClass2.enum1,
                    ]
                ]
            ),
            "enum_uint32.mat",
            "enum_uint32",
        ),
    ],
    ids=["enum-base", "enum-array", "enum-derived"],
)
def test_enum_instance(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert matdict[var_name].shape == expected_array.shape
    assert matdict[var_name].dtype == expected_array.dtype

    for idx in np.ndindex(expected_array.shape):
        actual_member = matdict[var_name][idx]
        expected_member = expected_array[idx]
        assert actual_member.name == expected_member.name
        assert actual_member.value == expected_member.value


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            {
                "_Class": "NestedClass",
                "_Props": np.array(
                    {
                        "objProp": np.array(
                            [
                                [
                                    EnumClass.enum1,
                                ]
                            ]
                        ),
                        "cellProp": np.array(
                            [
                                [
                                    np.array(
                                        [
                                            [
                                                EnumClass.enum2,
                                            ]
                                        ]
                                    )
                                ]
                            ],
                            dtype=object,
                        ),
                        "structProp": np.array(
                            [
                                [
                                    np.array(
                                        [
                                            [
                                                EnumClass.enum3,
                                            ]
                                        ]
                                    )
                                ]
                            ],
                            dtype=[("ObjField", "O")],
                        ),
                    }
                ).reshape(1, 1),
            },
            "enum_inside_obj.mat",
            "obj1",
        )
    ],
    ids=["enum-inside-object"],
)
def test_enum_inside_object(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict

    # Class Name
    assert matdict[var_name]["_Class"] == expected_array["_Class"]

    # Props Array
    assert matdict[var_name]["_Props"].shape == expected_array["_Props"].shape
    assert matdict[var_name]["_Props"].dtype == expected_array["_Props"].dtype

    # Props Dict
    actual_props = matdict[var_name]["_Props"][0, 0]
    expected_props = expected_array["_Props"][0, 0]
    for prop, val in expected_props.items():
        if prop == "cellProp":
            nested_actual_member = actual_props[prop][0, 0][0, 0]
            nested_expected_member = val[0, 0][0, 0]
            assert nested_actual_member.name == nested_expected_member.name
            assert nested_actual_member.value == nested_expected_member.value

        elif prop == "structProp":
            nested_actual_member = actual_props[prop]["ObjField"][0, 0][0, 0]
            nested_expected_member = val["ObjField"][0, 0][0, 0]
            assert nested_actual_member.name == nested_expected_member.name
            assert nested_actual_member.value == nested_expected_member.value

        elif prop == "objProp":
            nested_actual_member = actual_props[prop][0, 0]
            nested_expected_member = val[0, 0]
            assert nested_actual_member.name == nested_expected_member.name
            assert nested_actual_member.value == nested_expected_member.value



"""Conversion utilities for matio"""

from .matmap import mat_to_containermap, mat_to_dictionary
from .matstring import mat_to_string
from .mattables import mat_to_categorical, mat_to_table, mat_to_timetable
from .mattimes import mat_to_calendarduration, mat_to_datetime, mat_to_duration

__all__ = [
    "mat_to_containermap",
    "mat_to_dictionary",
    "mat_to_string",
    "mat_to_categorical",
    "mat_to_table",
    "mat_to_timetable",
    "mat_to_calendarduration",
    "mat_to_datetime",
    "mat_to_duration",
]



"""Utility functions for converting MATLAB containerMap and Dictionary"""

import warnings


def mat_to_containermap(props):
    """Converts the properties of a container map to a dictionary
    MATLAB container.map:
    - Property: serialization
        - Value: struct array
            - Fields: keys, values, uniformity, keyType, valueType (default= "any")
    """
    ks = props[0, 0]["serialization"][0, 0]["keys"]
    vals = props[0, 0]["serialization"][0, 0]["values"]

    result = {}
    for i in range(ks.shape[1]):
        key = ks[0, i].item()
        val = vals[0, i]
        result[key] = val

    return result


def mat_to_dictionary(props):
    """Converts the properties of a MATLAB dictionary to a dictionary
    MATLAB dictionary:
    - Property: data
        - Value: struct array
            - Fields: Version, IsKeyCombined, IsValueCombined, Key, Value

    Wrapped as tuple of (key, value) since keys can be of any type
    """
    ver = int(props[0, 0]["data"][0, 0]["Version"].item())
    if ver != 1:
        warnings.warn(
            f"Only v1 MATLAB dictionaries are supported. Got v{ver}",
            UserWarning,
        )

    ks = props[0, 0]["data"][0, 0]["Key"].ravel()
    vals = props[0, 0]["data"][0, 0]["Value"].ravel()
    return list(zip(ks, vals))



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            {
                "_Class": "NoConstructor",
                "_Props": np.array(
                    {
                        "a": np.array([]).reshape(0, 0),
                        "b": np.array([]).reshape(0, 0),
                        "c": np.array([]).reshape(0, 0),
                    }
                ).reshape(1, 1),
            },
            "object_without_constructor.mat",
            "obj1",
        ),
        (
            {
                "_Class": "YesConstructor",
                "_Props": np.array(
                    {
                        "a": np.array([10]).reshape(1, 1),
                        "b": np.array([20]).reshape(1, 1),
                        "c": np.array([30]).reshape(1, 1),
                    }
                ).reshape(1, 1),
            },
            "object_with_constructor.mat",
            "obj2",
        ),
        (
            {
                "_Class": "DefaultClass",
                "_Props": np.array(
                    {
                        "a": np.array([]).reshape(0, 0),
                        "b": np.array([10]).reshape(1, 1),
                        "c": np.array([30]).reshape(1, 1),
                    }
                ).reshape(1, 1),
            },
            "object_with_default.mat",
            "obj3",
        ),
        (
            {
                "_Class": "YesConstructor",
                "_Props": np.tile(
                    np.array(
                        {
                            "a": np.array([10]).reshape(1, 1),
                            "b": np.array([20]).reshape(1, 1),
                            "c": np.array([30]).reshape(1, 1),
                        }
                    ),
                    (2, 3),
                ),
            },
            "object_array.mat",
            "obj6",
        ),
    ],
    ids=[
        "object-without-constructor",
        "object-with-constructor",
        "object-with-default",
        "object-array",
    ],
)
def test_load_user_defined(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert matdict[var_name].keys() == expected_array.keys()

    # Class Name
    assert matdict[var_name]["_Class"] == expected_array["_Class"]

    # Property Dict
    assert matdict[var_name]["_Props"].shape == expected_array["_Props"].shape
    assert matdict[var_name]["_Props"].dtype == expected_array["_Props"].dtype

    # Each property, user-defined are stored as MxN arrays unlike MATLAB datatypes
    for idx in np.ndindex(expected_array["_Props"].shape):
        expected_props = expected_array["_Props"][idx]
        actual_props = matdict[var_name]["_Props"][idx]
        for prop, val in expected_props.items():
            np.testing.assert_array_equal(actual_props[prop], val)


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            {
                "_Class": "NestedClass",
                "_Props": np.array(
                    {
                        "objProp": {
                            "_Class": "NoConstructor",
                            "_Props": np.array(
                                {
                                    "a": np.array([]).reshape(0, 0),
                                    "b": np.array([]).reshape(0, 0),
                                    "c": np.array([]).reshape(0, 0),
                                }
                            ).reshape(1, 1),
                        },
                        "cellProp": np.array(
                            [
                                [
                                    {
                                        "_Class": "YesConstructor",
                                        "_Props": np.array(
                                            {
                                                "a": np.array([10]).reshape(1, 1),
                                                "b": np.array([20]).reshape(1, 1),
                                                "c": np.array([30]).reshape(1, 1),
                                            }
                                        ).reshape(1, 1),
                                    }
                                ]
                            ],
                            dtype=object,
                        ),
                        "structProp": np.array(
                            [
                                [
                                    {
                                        "_Class": "DefaultClass",
                                        "_Props": np.array(
                                            {
                                                "a": np.array([]).reshape(0, 0),
                                                "b": np.array([10]).reshape(1, 1),
                                                "c": np.array([30]).reshape(1, 1),
                                            }
                                        ).reshape(1, 1),
                                    }
                                ]
                            ],
                            dtype=[("ObjField", "O")],
                        ),
                    }
                ).reshape(1, 1),
            },
            "nested_object.mat",
            "obj4",
        )
    ],
    ids=["nested-object"],
)
def test_load_nested_user_defined(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert matdict[var_name].keys() == expected_array.keys()

    # Class Name
    assert matdict[var_name]["_Class"] == expected_array["_Class"]

    # Property Dict
    assert matdict[var_name]["_Props"].shape == expected_array["_Props"].shape
    assert matdict[var_name]["_Props"].dtype == expected_array["_Props"].dtype

    # Each Property
    actual_props = matdict[var_name]["_Props"][0, 0]
    expected_props = expected_array["_Props"][0, 0]
    for prop, val in expected_props.items():
        if prop == "cellProp":
            nested_actual_dict = actual_props[prop][0, 0]
            nested_expected_dict = val[0, 0]
            assert nested_actual_dict["_Class"] == nested_expected_dict["_Class"]
            assert (
                nested_actual_dict["_Props"].shape
                == nested_expected_dict["_Props"].shape
            )
            assert (
                nested_actual_dict["_Props"].dtype
                == nested_expected_dict["_Props"].dtype
            )
            nested_expected_props = nested_expected_dict["_Props"][0, 0]
            nested_actual_props = nested_actual_dict["_Props"][0, 0]
            for prop, val in nested_expected_props.items():
                if isinstance(val, np.ndarray):
                    np.testing.assert_array_equal(nested_actual_props[prop], val)
                else:
                    assert nested_actual_props[prop] == val
        elif prop == "structProp":
            nested_actual_dict = actual_props[prop]["ObjField"][0, 0]
            nested_expected_dict = val["ObjField"][0, 0]
            assert nested_actual_dict["_Class"] == nested_expected_dict["_Class"]
            assert (
                nested_actual_dict["_Props"].shape
                == nested_expected_dict["_Props"].shape
            )
            assert (
                nested_actual_dict["_Props"].dtype
                == nested_expected_dict["_Props"].dtype
            )
            nested_expected_props = nested_expected_dict["_Props"][0, 0]
            nested_actual_props = nested_actual_dict["_Props"][0, 0]
            for prop, val in nested_expected_props.items():
                if isinstance(val, np.ndarray):
                    np.testing.assert_array_equal(nested_actual_props[prop], val)
                else:
                    assert nested_actual_props[prop] == val

        elif prop == "objProp":
            nested_actual_dict = actual_props[prop]
            nested_expected_dict = val

            assert nested_actual_dict["_Class"] == nested_expected_dict["_Class"]
            assert (
                nested_actual_dict["_Props"].shape
                == nested_expected_dict["_Props"].shape
            )
            assert (
                nested_actual_dict["_Props"].dtype
                == nested_expected_dict["_Props"].dtype
            )

            nested_expected_props = nested_expected_dict["_Props"][0, 0]
            nested_actual_props = nested_actual_dict["_Props"][0, 0]
            for prop, val in nested_expected_props.items():
                if isinstance(val, np.ndarray):
                    np.testing.assert_array_equal(nested_actual_props[prop], val)
                else:
                    assert nested_actual_props[prop] == val


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            {
                "_Class": "DefaultClass2",
                "_Props": np.array(
                    {
                        "a": np.array(["Default String"]).reshape(1, 1),
                        "b": np.array([10]).reshape(1, 1),
                        "c": np.array([30]).reshape(1, 1),
                    }
                ).reshape(1, 1),
            },
            "object_in_default_prop.mat",
            "obj7",
        ),
    ],
    ids=["object-in-default-property"],
)
def test_load_user_defined_with_default_property(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    # Output format
    assert var_name in matdict
    assert matdict[var_name].keys() == expected_array.keys()

    # Class Name
    assert matdict[var_name]["_Class"] == expected_array["_Class"]

    # Property Dict
    assert matdict[var_name]["_Props"].shape == expected_array["_Props"].shape
    assert matdict[var_name]["_Props"].dtype == expected_array["_Props"].dtype

    # Each property, user-defined are stored as MxN arrays unlike MATLAB datatypes
    for idx in np.ndindex(expected_array["_Props"].shape):
        expected_props = expected_array["_Props"][idx]
        actual_props = matdict[var_name]["_Props"][idx]
        for prop, val in expected_props.items():
            if isinstance(val, np.ndarray):
                np.testing.assert_array_equal(actual_props[prop], val)
            elif isinstance(val, dict):
                nested_actual_dict = actual_props[prop]
                nested_expected_dict = val

                assert nested_actual_dict["_Class"] == nested_expected_dict["_Class"]
                assert (
                    nested_actual_dict["_Props"].shape
                    == nested_expected_dict["_Props"].shape
                )
                assert (
                    nested_actual_dict["_Props"].dtype
                    == nested_expected_dict["_Props"].dtype
                )

                nested_expected_props = nested_expected_dict["_Props"][0, 0]
                nested_actual_props = nested_actual_dict["_Props"][0, 0]
                for prop, val in nested_expected_props.items():
                    np.testing.assert_array_equal(nested_actual_props[prop], val)



import os

import numpy as np
import pytest

from matio import load_from_mat


@pytest.mark.parametrize(
    "expected_array, file_name, var_name",
    [
        (
            np.array(["Hello"], dtype=np.str_).reshape(1, 1),
            "string_base.mat",
            "s1",
        ),
        (
            np.array(
                ["Apple", "Banana", "Cherry", "Date", "Fig", "Grapes"], dtype=np.str_
            ).reshape(2, 3),
            "string_array.mat",
            "s2",
        ),
        (
            np.array([""], dtype=np.str_).reshape(1, 1),
            "string_empty.mat",
            "s3",
        ),
    ],
    ids=["simple-string", "string-array", "empty-string"],
)
def test_parse_string(expected_array, file_name, var_name):
    file_path = os.path.join(os.path.dirname(__file__), file_name)
    matdict = load_from_mat(file_path, raw_data=False)

    assert var_name in matdict
    np.testing.assert_array_equal(matdict[var_name], expected_array)



"""Utility functions for converting MATLAB datetime, duration, and calendarDuration"""

import warnings
from datetime import datetime
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import numpy as np


def get_tz_offset(tz):
    """Get timezone offset in milliseconds
    Inputs:
        1. tz (str): Timezone string
    Returns:
        1. offset (int): Timezone offset in milliseconds
    """
    try:
        tzinfo = ZoneInfo(tz)
        utc_offset = tzinfo.utcoffset(datetime.now())
        if utc_offset is not None:
            offset = int(utc_offset.total_seconds() * 1000)
        else:
            offset = 0
    except ZoneInfoNotFoundError as e:
        warnings.warn(
            f"Could not get timezone offset for {tz}: {e}. Defaulting to UTC."
        )
        offset = 0
    return offset


def mat_to_datetime(props):
    """Convert MATLAB datetime to Python datetime
    Datetime returned as numpy.datetime64[ms]

    MATLAB datetimes objects are stored with the following properties:
    1. data - complex number = real_part (ms) + i * imag_part (us)
    2. fmt - Format | char array
    3. tz - Timezone | char array
    """

    data = props[0, 0].get("data", np.array([]))
    if data.size == 0:
        return np.array([], dtype="datetime64[ms]")
    tz = props[0, 0].get("tz", None)
    if tz is not None and tz.size > 0:
        offset = get_tz_offset(tz.item())
    else:
        offset = 0

    millis = data.real + data.imag * 1e3 + offset

    return millis.astype("datetime64[ms]")


def mat_to_duration(props):
    """Convert MATLAB duration to Python timedelta
    Duration returned as numpy.timedelta64

    MATLAB datetimes objects are stored with the following properties:
    1. millis - double
    2. fmt - char array
    """

    millis = props[0, 0]["millis"]
    if millis.size == 0:
        return np.array([], dtype="timedelta64[ms]")

    fmt = props[0, 0].get("fmt", None)
    if fmt is None:
        return millis.astype("timedelta64[ms]")

    if fmt == "s":
        count = millis / 1000  # Seconds
        dur = count.astype("timedelta64[s]")
    elif fmt == "m":
        count = millis / (1000 * 60)  # Minutes
        dur = count.astype("timedelta64[m]")
    elif fmt == "h":
        count = millis / (1000 * 60 * 60)  # Hours
        dur = count.astype("timedelta64[h]")
    elif fmt == "d":
        count = millis / (1000 * 60 * 60 * 24)  # Days
        dur = count.astype("timedelta64[D]")
    elif fmt == "y":
        count = millis / (1000 * 60 * 60 * 24 * 365)  # Years
        dur = count.astype("timedelta64[Y]")
    else:
        count = millis
        dur = count.astype("timedelta64[ms]")
        # Default case

    return dur


def mat_to_calendarduration(props):
    """Convert MATLAB calendarDuration to Python timedelta
    CalendarDuration returned as numpy.timedelta64

    MATLAB calendarDuration objects are stored with the following properties:
    1. months - double
    2. days - double
    3. millis - double
    """

    months = props[0, 0]["components"][0, 0]["months"].astype("timedelta64[M]")
    days = props[0, 0]["components"][0, 0]["days"].astype("timedelta64[D]")
    millis = props[0, 0]["components"][0, 0]["millis"].astype("timedelta64[ms]")

    cal = np.empty((1, 1), dtype=[("calendarDuration", object)])
    cal[0, 0]["calendarDuration"] = (months, days, millis)
    return cal
